from pydantic import AnyUrl, BaseModel, ConfigDict
from pydantic_settings import BaseSettings


class DomainControllersSettings(BaseModel):
    primary_uri: AnyUrl
    secondary_uri: AnyUrl | None = None


class UsersSettings(BaseModel):
    dns: list[str] | None = None


class ActiveDirectorySettings(BaseSettings):
    base_dn: str = ""
    bind_dn: str = ""
    bind_password: str = ""
    domain_controllers: DomainControllersSettings
    users: UsersSettings | None = None


class IntegrationTestData(BaseModel):
    model_config = ConfigDict(extra="allow")

    base_dn: str | None = None
    attributes: list[dict[str, str]] | None = None


class IntegrationTestsSections(BaseModel):
    model_config = ConfigDict(extra="allow")

    users: IntegrationTestData | None = None
    get_user: IntegrationTestData | None = None
    list_users: IntegrationTestData | None = None
    list_groups: IntegrationTestData | None = None



from typing import Any, Literal

import ldap
from ldap.ldapobject import LDAPObject

from . import utils
from .exceptions import ADError, UserDoesNotFound
from .models import ADGroup, ADGroups, ADUser, ADUsers, SearchRawOut, ADObject
from .settings import ActiveDirectorySettings

LdapScope = Literal["onelevel", "base", "subtree"]


class ActiveDirectoryConnector:
    def __init__(self, settings: ActiveDirectorySettings):
        self.connection: LDAPObject | None = None
        self.settings = settings

    @classmethod
    def get_scope(cls, scope: LdapScope) -> int:
        # scope: one_level [SCOPE_ONELEVEL] base [SCOPE_BASE] subtree [SCOPE_SUBTREE]
        ldap_scope = getattr(ldap, f"SCOPE_{scope.upper()}", ldap.SCOPE_ONELEVEL)  # pyright: ignore
        return ldap_scope

    def bind(self, ldap_options: dict[Any, Any] | None = None):
        ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)  # pyright: ignore
        if ldap_options:
            for (
                option,
                value,
            ) in ldap_options.items():
                ldap.set_option(option, value)

        self.connection: LDAPObject = ldap.initialize(
            str(self.settings.domain_controllers.primary_uri)
        )
        self.connection.start_tls_s()
        self.connection.simple_bind_s(self.settings.bind_dn, self.settings.bind_password)

    def search(
        self,
        to_search: str,
        base_dn: str = "",
        scope: LdapScope | None = None,
        attributes: list | None = None,
    ) -> SearchRawOut:
        base_dn = base_dn or self.settings.base_dn
        scope = scope if scope else "onelevel"
        results = self.connection.search_s(
            base_dn, self.get_scope(scope=scope), to_search, attributes
        )
        return SearchRawOut(results)

    def list_objects(
        self,
        out_object: ADObject,
        to_search: str | None = None,
        base_dn: str = "",
        scope: LdapScope | None = None,
        attributes: list[str] | None = None,
    ) -> ADUsers | ADGroups:
        attributes = attributes or out_object.attributes()
        if "sAMAccountName" not in attributes:
            attributes += ["sAMAccountName"]

        ADObjectOut = None
        if out_object == ADUser:
            to_search = utils.FILTER_SEARCH_USERS
            ADObjectOut = ADUsers
        elif out_object == ADGroup:
            to_search = utils.FILTER_SEARCH_GROUPS
            ADObjectOut = ADGroups
        else:
            raise Exception(f"Object class {out_object} not supported")

        if not to_search:
            raise Exception("Specificare filter search")
        try:
            results = self.search(
                to_search=to_search,
                base_dn=base_dn,
                scope=scope,
                attributes=attributes,
            )
        except ldap.NO_SUCH_OBJECT:
            return ADObjectOut()
        except ldap.LDAPError as exc:
            raise ADError(ldap_error=exc) from exc
        objects = {data[1]["sAMAccountName"][0].lower(): out_object(**data[1]) for data in results}
        return ADObjectOut(objects)

    def list_users(
        self, base_dn: str = "", scope: LdapScope | None = None, attributes: list[str] | None = None
    ) -> ADUsers:
        return self.list_objects(
            out_object=ADUser, base_dn=base_dn, scope=scope, attributes=attributes
        )

    def list_groups(
        self, base_dn: str = "", scope: LdapScope | None = None, attributes: list[str] | None = None
    ) -> ADUsers:
        return self.list_objects(
            out_object=ADGroup, base_dn=base_dn, scope=scope, attributes=attributes
        )

    def get_user(
        self,
        username: str,
        base_dn: str = "",
        scope: LdapScope | None = None,
        attributes: list | None = None,
    ) -> ADUser:
        to_search = utils.FILTER_SEARCH_USER_BY_SAMACC.format(username=username)
        attributes = attributes or ADUser.attributes()
        try:
            results = self.search(
                to_search=to_search, base_dn=base_dn, scope=scope, attributes=attributes
            )
        except ldap.LDAPError as exc:
            raise ADError(ldap_error=exc) from exc
        if len(results) == 0:
            raise UserDoesNotFound(f"User '{username}' not found in '{base_dn}'")
        return ADUser(**results[0][1])

    def get_domain_data(self, base_dn: str = "", attributes=None) -> SearchRawOut:
        base_dn = base_dn or self.settings.base_dn
        try:
            return self.search(
                to_search="(objectCategory=domain)",
                base_dn=base_dn,
                scope="subtree",  # pylint: disable=no-member # pyright: ignore
                attributes=attributes,
            )
        except ldap.LDAPError as exc:
            raise ADError(ldap_error=exc) from exc

    def disable_user(self, user_dn: str) -> None:
        mod_list_attrs = [(ldap.MOD_REPLACE, "userAccountControl", b"514")]  # pyright: ignore
        res = self.connection.modify_s(user_dn, mod_list_attrs)
        return res



class ADError(Exception):
    def __init__(self, ldap_error: Exception | None = None):
        super().__init__(ldap_error)
        self.ldap_error = ldap_error


class ObjectDoesNotFound(ADError):
    pass


class UserDoesNotFound(ObjectDoesNotFound):
    pass



import platform
import struct
import subprocess
from datetime import datetime, timedelta

import pytz

FILTER_SEARCH_USERS = "(&(objectCategory=person)(objectclass=user))"
FILTER_SEARCH_USER_BY_SAMACC = FILTER_SEARCH_USERS[:-1] + "(sAMAccountName={username}))"
FILTER_SEARCH_GROUPS = "(&(objectclass=group))"
FILTER_SEARCH_GROUP_BY_DN = "(&(objectclass=group)(dn={dn}))"


def sid_to_str(sid: bytes) -> str:
    """Converts a hexadecimal string returned from the LDAP query to a
    string version of the SID in format of S-1-5-21-1270288957-3800934213-3019856503-500

    ref: https://gist.github.com/mprahl/e38a2eba6da09b2f6bd69d30fd3b749e
    """
    revision = sid[0]
    number_of_sub_ids = sid[1]
    iav = struct.unpack(">Q", b"\x00\x00" + sid[2:8])[0]
    sub_ids = [
        struct.unpack("<I", sid[8 + 4 * i : 12 + 4 * i])[0] for i in range(number_of_sub_ids)
    ]
    return "S-{}-{}-{}".format(revision, iav, "-".join([str(sub_id) for sub_id in sub_ids]))


def winntformat_to_datetime(winntformat, timezone: str = "Europe/Rome") -> datetime:
    timestamp = float(winntformat)
    seconds_since_epoch = timestamp / 10**7
    loc_dt = datetime.fromtimestamp(seconds_since_epoch)
    loc_dt -= timedelta(days=(1970 - 1601) * 365 + 89)
    return pytz.timezone(timezone).localize(loc_dt)


def ping(host=None, retry_packets=1):
    param = "-n" if platform.system().lower() == "windows" else "-c"
    command = ["ping", param, str(retry_packets), host]
    return subprocess.call(command) == 0
